import 'dart:core';
import 'dart:math';

class DataCapsule {
  // * Public methods starts here

  /// Generate a prescription key to be used for packing and unpacking of data capsules.
  String prescription() {
    List _initialPattern = List.from(this._generateInitialPattern().codeUnits);
    var _randomHandler = Random();
    for (int _maxReps = 0; _maxReps < 5; _maxReps++)
      _initialPattern.shuffle(_randomHandler);
    StringBuffer _assembleString = StringBuffer();
    for (int _pattern in _initialPattern)
      _assembleString.write(String.fromCharCode(_pattern));
    return _assembleString.toString();
  }

  /// Pack a single string data into a data capsule using a generated prescription key.
  String pack(String data, String prescription) {
    List<String> _catalyst = this._generateCatalyst(prescription);
    return _pack(_catalyst, data);
  }

  /// Pack a list of string data into data capsules using a generated prescription key.
  List<String> packList(List<String> datalist, String prescription) {
    List<String> _catalyst = this._generateCatalyst(prescription);
    for (String _data in datalist) _data = this._pack(_catalyst, _data);
    return datalist;
  }

  /// Unpack a single string data using the same key that was used for packaging.
  String unpack(String capsule, String prescription) {
    List<String> _catalyst = this._generateCatalyst(prescription);
    return _unpack(_catalyst, capsule);
  }

  /// Unpack a list string data using the same key that was used for packaging.
  List<String> unpackList(List<String> capsules, String prescription) {
    List<String> _catalyst = this._generateCatalyst(prescription);
    for (String _capsule in capsules)
      _capsule = this._unpack(_catalyst, _capsule);
    return capsules;
  }

  // * Public methods ends here

  //* Private methods starts here

  String _generateInitialPattern() {
    StringBuffer _asciiArray = StringBuffer();
    for (int i = 65; i <= 90; i++) _asciiArray.write(String.fromCharCode(i));
    for (int i = 0; i < 10; i++)
      if (![0, 2, 1, 4].contains(i)) _asciiArray.write(i.toString());
    return _asciiArray.toString();
  }

  bool _isPrescriptionValid(String prescription) {
    bool _error = false;
    List _initialPattern = this._generateInitialPattern().codeUnits;
    for (int _key in prescription.codeUnits)
      _error = !_error ? !_initialPattern.contains(_key) : !_error;
    return prescription.codeUnits.length == _initialPattern.length && !_error;
  }

  List<String> _generateCatalyst(String prescription) {
    List<String> _catalysts = [];
    StringBuffer _stringBuffer = StringBuffer();
    if (this._isPrescriptionValid(prescription)) {
      for (int _catalyst = 1; _catalyst <= 4; _catalyst++) {
        _stringBuffer.clear();
        for (int _key in prescription.codeUnits) {
          if (_key >= 65 && _key <= 90) {
            _stringBuffer.write(String.fromCharCode(
                (_key + (_catalyst * 4)) > 90
                    ? 64 + ((_key + (_catalyst * 4)) - 90)
                    : (_key + (_catalyst * 4))));
          }
        }
        _catalysts.add(_stringBuffer.toString());
      }
    }
    _catalysts[1] = this._modifyCatalyst(_catalysts[1]);
    _catalysts[2] = this._modifyCatalyst(_catalysts[2]);
    return _catalysts;
  }

  String _modifyCatalyst(String catalyst) {
    String _newCatalyst = '';
    StringBuffer _firstHalf = StringBuffer(), _secondHalf = StringBuffer();

    // Reverse the first half of the catalyst string
    for (int i = 0; i < catalyst.codeUnits.length / 2; i++) {
      _firstHalf.write(String.fromCharCode(catalyst.codeUnits[i]));
    }
    for (int key in _firstHalf.toString().codeUnits.reversed.toList()) {
      _newCatalyst = _newCatalyst + String.fromCharCode(key);
    }

    // Reverse the second half of the catalyst string
    for (int i = 13; i < catalyst.codeUnits.length; i++) {
      _secondHalf.write(String.fromCharCode(catalyst.codeUnits[i]));
    }
    for (int key in _secondHalf.toString().codeUnits.reversed.toList()) {
      _newCatalyst = _newCatalyst + String.fromCharCode(key);
    }

    return _newCatalyst;
  }

  List<String> _lowerCaseCatalyst(List<String> catalyst) {
    List<String> _lowerCase = [];
    for (String _str in catalyst) _lowerCase.add(_str.toLowerCase());
    return _lowerCase;
  }

  String _pack(List<String> catalyst, String data) {
    if (catalyst.isEmpty)
      throw Exception("Catalyst was not created correctly.");
    StringBuffer _capsuleBuilder = StringBuffer();
    List<String> lowerCatalyst = this._lowerCaseCatalyst(catalyst);

    int _cipherSwitcher = 0;
    for (int key in data.codeUnits) {
      if ((key >= 65 && key <= 90) || (key >= 97 && key <= 122)) {
        if (key >= 65 && key <= 90) {
          int _index =
              catalyst[_cipherSwitcher].indexOf(String.fromCharCode(key), 0);
          int _key = catalyst[_cipherSwitcher + (_cipherSwitcher > 2 ? -2 : 2)]
              .codeUnitAt(_index);
          _capsuleBuilder.writeCharCode(_key);
        } else if (key >= 97 && key <= 122) {
          int _index = lowerCatalyst[_cipherSwitcher]
              .indexOf(String.fromCharCode(key), 0);
          int _key =
              lowerCatalyst[_cipherSwitcher + (_cipherSwitcher > 2 ? -2 : 2)]
                  .codeUnitAt(_index);
          _capsuleBuilder.writeCharCode(_key);
        }
        _cipherSwitcher = _cipherSwitcher + (_cipherSwitcher == 0 ? 3 : -3);
      } else
        _capsuleBuilder.writeCharCode(key);
    }
    return _capsuleBuilder.toString();
  }

  String _unpack(List<String> catalyst, String data) {
    if (catalyst.isEmpty)
      throw Exception("Catalyst was not created correctly.");
    StringBuffer _dataBuilder = StringBuffer();
    List<String> lowerCatalyst = this._lowerCaseCatalyst(catalyst);

    int _cipherSwitcher = 2;
    for (int key in data.codeUnits) {
      if ((key >= 65 && key <= 90) || (key >= 97 && key <= 122)) {
        if (key >= 65 && key <= 90) {
          int _index =
              catalyst[_cipherSwitcher].indexOf(String.fromCharCode(key), 0);
          int _key = catalyst[_cipherSwitcher + (_cipherSwitcher == 2 ? -2 : 2)]
              .codeUnitAt(_index);
          _dataBuilder.writeCharCode(_key);
        } else if (key >= 97 && key <= 122) {
          int _index = lowerCatalyst[_cipherSwitcher]
              .indexOf(String.fromCharCode(key), 0);
          int _key =
              lowerCatalyst[_cipherSwitcher + (_cipherSwitcher == 2 ? -2 : 2)]
                  .codeUnitAt(_index);
          _dataBuilder.writeCharCode(_key);
        }
        _cipherSwitcher = _cipherSwitcher + (_cipherSwitcher == 2 ? -1 : 1);
      } else
        _dataBuilder.writeCharCode(key);
    }
    return _dataBuilder.toString();
  }

  //* Private methods ends here
}
